@@ -222,6 +222,8 @@ if ($^O eq 'openbsd' && $args{jit} && $config{cc} eq 'clang') {
 # Set the remaining ldmiscflags. Do this after probing for gcc -Werror probe to not miss that change for the linker.
 $config{ldmiscflags}  = $config{ccmiscflags} unless defined $config{ldmiscflags};
 
+# Include paths that NQP/Rakudo are going to need in their build.
+my @hllincludes = qw( moar );
 
 if ($args{'has-sha'}) {
     $config{shaincludedir} = '/usr/include/sha';
@@ -230,7 +232,7 @@ if ($args{'has-sha'}) {
 }
 else { $config{shaincludedir} = '3rdparty/sha1' }
 
-# After upgrading from libuv from 0.11.18 to 0.11.29 we see very weird erros
+# After upgrading from libuv from 0.11.18 to 0.11.29 we see very weird errors
 # when the old libuv files are still around. Running a `make realclean` in
 # case we spot an old file and the Makefile is already there.
 if (-e '3rdparty/libuv/src/unix/threadpool' . $defaults{obj}
@@ -268,6 +270,7 @@ else {
                         . "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/libuv/uv\"\n"
                         . "\t\$(CP) 3rdparty/libuv/include/*.h \"\$(DESTDIR)\$(PREFIX)/include/libuv\"\n"
                         . "\t\$(CP) 3rdparty/libuv/include/uv/*.h \"\$(DESTDIR)\$(PREFIX)/include/libuv/uv\"\n";
+    push @hllincludes, 'libuv';
 }
 
 if ($args{'has-libatomic_ops'}) {
@@ -297,6 +300,7 @@ else {
                         . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/loadstore/*.h \"$lao/atomic_ops/sysdeps/loadstore\"\n"
                         . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/msftc/*.h \"$lao/atomic_ops/sysdeps/msftc\"\n"
                         . "\t\$(CP) 3rdparty/libatomicops/src/atomic_ops/sysdeps/sunc/*.h \"$lao/atomic_ops/sysdeps/sunc\"\n";
+    push @hllincludes, 'libatomic_ops';
 }
 
 if ($args{'has-libtommath'}) {
@@ -315,6 +319,7 @@ else {
     $config{moar_cincludes} .= ' ' . $defaults{ccinc} . '3rdparty/libtommath';
     $config{install}   .= "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/libtommath\"\n"
                         . "\t\$(CP) 3rdparty/libtommath/*.h \"\$(DESTDIR)\$(PREFIX)/include/libtommath\"\n";
+    push @hllincludes, 'libtommath';
 }
 
 if ($args{'has-libffi'}) {
@@ -362,6 +367,7 @@ else {
                         . "\t\$(CP) 3rdparty/dyncall/dynload/*.h \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n"
                         . "\t\$(CP) 3rdparty/dyncall/dyncall/*.h \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n"
                         . "\t\$(CP) 3rdparty/dyncall/dyncallback/*.h \"\$(DESTDIR)\$(PREFIX)/include/dyncall\"\n";
+    push @hllincludes, 'dyncall';
 }
 
 # The ZSTD_CStream API is only exposed starting at version 1.0.0
@@ -432,6 +438,17 @@ push @cflags, '-DWORDS_BIGENDIAN' if $config{be}; # 3rdparty/sha1 needs it and i
 push @cflags, '-DMVM_HEAPSNAPSHOT_FORMAT=' . $config{heapsnapformat};
 push @cflags, $ENV{CFLAGS} if $ENV{CFLAGS};
 push @cflags, $ENV{CPPFLAGS} if $ENV{CPPFLAGS};
+
+# Define _GNU_SOURCE for libuv to quell warnings with gcc.
+# According to the libuv developers, they define _GNU_SOURCE for
+# Linux use. Our code was showing warnings from libuv
+# because we did not. Adding -D_GNU_SOURCE conditionally
+# to our build then resulted in redefinition warnings in
+# our code for three of our files that already have that definition.
+# The fix was to bracket those definitions with
+# #ifdef _GNU_SOURCE/#endif.
+push @cflags, '-D_GNU_SOURCE' unless $args{'has-libuv'};
+
 $config{cflags} = join ' ', uniq(@cflags);
 
 # generate LDFLAGS
@@ -577,12 +594,6 @@ unless (defined $config{jit_obj}) {
     $config{dasm_flags}   = '';
 }
 
-
-if ($config{cc} eq 'cl') {
-    $config{install}   .= "\t\$(MKPATH) \"\$(DESTDIR)\$(PREFIX)/include/msinttypes\"\n"
-                        . "\t\$(CP) 3rdparty/msinttypes/*.h \"\$(DESTDIR)\$(PREFIX)/include/msinttypes\"\n";
-}
-
 if ($^O eq 'aix' && $config{ptr_size} == 4) {
     $config{ldflags} = join(',', $config{ldflags}, '-bmaxdata:0x80000000');
 }
@@ -675,6 +686,10 @@ my $thirdpartylibs = join "\n" . ' ' x 12, sort @thirdpartylibs;
 
 print "OK\n";
 
+# the parser for config values used in the NQP/Rakudo build doesn't understand arrays,
+# so just join into a single string
+$config{hllincludes} = join " ", @hllincludes;
+
 write_backend_config();
 
 # dump 3rdparty libs we need to build
