@@ -33,17 +33,23 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
         if (ss->boxed_primitive == MVM_STORAGE_SPEC_BP_INT) {
             if (ss->bits == 8 || ss->bits == 16 || ss->bits == 32 || ss->bits == 64)
                 repr_data->elem_size = ss->bits / 8;
-            else
+            else {
+                MVM_free(repr_data);
+                st->REPR_data = NULL;
                 MVM_exception_throw_adhoc(tc,
                     "CArray representation can only have 8, 16, 32 or 64 bit integer elements");
+            }
             repr_data->elem_kind = MVM_CARRAY_ELEM_KIND_NUMERIC;
         }
         else if (ss->boxed_primitive == MVM_STORAGE_SPEC_BP_NUM) {
             if (ss->bits == 32 || ss->bits == 64)
                 repr_data->elem_size = ss->bits / 8;
-            else
+            else {
+                MVM_free(repr_data);
+                st->REPR_data = NULL;
                 MVM_exception_throw_adhoc(tc,
                     "CArray representation can only have 32 or 64 bit floating point elements");
+            }
             repr_data->elem_kind = MVM_CARRAY_ELEM_KIND_NUMERIC;
         }
         else if (ss->can_box & MVM_STORAGE_SPEC_CAN_BOX_STR) {
@@ -58,11 +64,21 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
             repr_data->elem_kind = MVM_CARRAY_ELEM_KIND_CSTRUCT;
             repr_data->elem_size = sizeof(void *);
         }
+        else if (type_id == MVM_REPR_ID_MVMCPPStruct) {
+            repr_data->elem_kind = MVM_CARRAY_ELEM_KIND_CPPSTRUCT;
+            repr_data->elem_size = sizeof(void *);
+        }
+        else if (type_id == MVM_REPR_ID_MVMCUnion) {
+            repr_data->elem_kind = MVM_CARRAY_ELEM_KIND_CUNION;
+            repr_data->elem_size = sizeof(void *);
+        }
         else if (type_id == MVM_REPR_ID_MVMCPointer) {
             repr_data->elem_kind = MVM_CARRAY_ELEM_KIND_CPOINTER;
             repr_data->elem_size = sizeof(void *);
         }
         else {
+            MVM_free(repr_data);
+            st->REPR_data = NULL;
             MVM_exception_throw_adhoc(tc,
                 "CArray representation only handles attributes of type:\n"
                 "  (u)int8, (u)int16, (u)int32, (u)int64, (u)long, (u)longlong, num32, num64, (s)size_t, bool, Str\n"
@@ -119,10 +135,18 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 /* This is called to do any cleanup of resources when an object gets
  * embedded inside another one. Never called on a top-level object. */
 static void gc_cleanup(MVMThreadContext *tc, MVMSTable *st, void *data) {
-    MVMCArrayBody *body = (MVMCArrayBody *)data;
+    MVMCArrayBody     *body      = (MVMCArrayBody *)data;
+    MVMCArrayREPRData *repr_data = (MVMCArrayREPRData *)st->REPR_data;
+    MVMint32 i;
 
-    if (body->managed) 
+    if (body->managed) {
+        if (repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_STRING) {
+            for (i = 0; i < body->elems; i++) {
+                MVM_free( ((void **)body->storage)[i] );
+            }
+        }
         MVM_free(body->storage);
+    }
     if (body->child_objs)
         MVM_free(body->child_objs);
 }
@@ -172,7 +196,7 @@ static const MVMStorageSpec * get_storage_spec(MVMThreadContext *tc, MVMSTable *
 }
 
 
-MVM_NO_RETURN static void die_pos_nyi(MVMThreadContext *tc) MVM_NO_RETURN_GCC;
+MVM_NO_RETURN static void die_pos_nyi(MVMThreadContext *tc) MVM_NO_RETURN_ATTRIBUTE;
 static void die_pos_nyi(MVMThreadContext *tc) {
     MVM_exception_throw_adhoc(tc,
         "CArray representation does not fully positional storage yet");
@@ -190,21 +214,21 @@ static void expand(MVMThreadContext *tc, MVMCArrayREPRData *repr_data, MVMCArray
         const size_t old_size = body->allocated * repr_data->elem_size;
         const size_t new_size = next_size * repr_data->elem_size;
 
-        body->storage = MVM_realloc(body->storage, new_size);
-        memset((char *)body->storage + old_size, 0, new_size - old_size);
+        body->storage = MVM_recalloc(body->storage, old_size, new_size);
     }
 
     is_complex = (repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CARRAY
                || repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CPOINTER
                || repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CSTRUCT
+               || repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CPPSTRUCT
+               || repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CUNION
                || repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_STRING);
 
     if (is_complex) {
         const size_t old_size = body->allocated * sizeof(MVMObject *);
         const size_t new_size = next_size * sizeof(MVMObject *);
 
-        body->child_objs = (MVMObject **) MVM_realloc(body->child_objs, new_size);
-        memset((char *)body->child_objs + old_size, 0, new_size - old_size);
+        body->child_objs = (MVMObject **) MVM_recalloc(body->child_objs, old_size, new_size);
     }
 
     body->allocated = next_size;
@@ -371,6 +395,18 @@ static void bind_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void
             bind_wrapper_and_ptr(tc, root, body, index, value.o,
                 IS_CONCRETE(value.o) ? ((MVMCStruct *)value.o)->body.cstruct : NULL);
             break;
+        case MVM_CARRAY_ELEM_KIND_CPPSTRUCT:
+            if (REPR(value.o)->ID != MVM_REPR_ID_MVMCPPStruct)
+                MVM_exception_throw_adhoc(tc, "CArray of CPPStruct passed non-CStruct object");
+            bind_wrapper_and_ptr(tc, root, body, index, value.o,
+                IS_CONCRETE(value.o) ? ((MVMCPPStruct *)value.o)->body.cppstruct : NULL);
+            break;
+        case MVM_CARRAY_ELEM_KIND_CUNION:
+            if (REPR(value.o)->ID != MVM_REPR_ID_MVMCUnion)
+                MVM_exception_throw_adhoc(tc, "CArray of CUnion passed non-CStruct object");
+            bind_wrapper_and_ptr(tc, root, body, index, value.o,
+                IS_CONCRETE(value.o) ? ((MVMCUnion *)value.o)->body.cunion : NULL);
+            break;
         default:
             MVM_exception_throw_adhoc(tc, "Unknown element type in CArray");
     }
@@ -392,6 +428,10 @@ static void shift(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *da
     die_pos_nyi(tc);
 }
 
+static void aslice(MVMThreadContext *tc, MVMSTable *st, MVMObject *src, void *data, MVMObject *dest, MVMint64 start, MVMint64 end) {
+    die_pos_nyi(tc);
+}
+
 static MVMuint64 elems(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data) {
     MVMCArrayBody *body = (MVMCArrayBody *)data;
 
@@ -402,6 +442,21 @@ static MVMuint64 elems(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, voi
         "Don't know how many elements a C array returned from a library");
 }
 
+static MVMuint64 unmanaged_size(MVMThreadContext *tc, MVMSTable *st, void *data) {
+    MVMCArrayREPRData *repr_data = (MVMCArrayREPRData *)st->REPR_data;
+    MVMCArrayBody     *body      = (MVMCArrayBody *)data;
+    MVMuint64 result = 0;
+
+    /* The allocated (or just-pointed-at) memory block */
+    result += body->allocated * repr_data->elem_size;
+
+    /* The array we hold wrapper objects in */
+    if (body->child_objs)
+        result += body->allocated * sizeof(MVMObject *);
+
+    return result;
+}
+
 static void deserialize_stable_size(MVMThreadContext *tc, MVMSTable *st, MVMSerializationReader *reader) {
     st->size = sizeof(MVMCArray);
 }
@@ -455,6 +510,7 @@ static const MVMREPROps CArray_this_repr = {
         pop,
         unshift,
         shift,
+        aslice,
         MVM_REPR_DEFAULT_SPLICE,
         MVM_REPR_DEFAULT_AT_POS_MULTIDIM,
         MVM_REPR_DEFAULT_BIND_POS_MULTIDIM,
@@ -462,7 +518,9 @@ static const MVMREPROps CArray_this_repr = {
         MVM_REPR_DEFAULT_SET_DIMENSIONS,
         MVM_REPR_DEFAULT_GET_ELEM_STORAGE_SPEC,
         MVM_REPR_DEFAULT_POS_AS_ATOMIC,
-        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM
+        MVM_REPR_DEFAULT_POS_AS_ATOMIC_MULTIDIM,
+        MVM_REPR_DEFAULT_POS_WRITE_BUF,
+        MVM_REPR_DEFAULT_POS_READ_BUF
     },    /* pos_funcs */
     MVM_REPR_DEFAULT_ASS_FUNCS,
     elems,
@@ -482,6 +540,6 @@ static const MVMREPROps CArray_this_repr = {
     NULL, /* spesh */
     "CArray", /* name */
     MVM_REPR_ID_MVMCArray,
-    NULL, /* unmanaged_size */
+    unmanaged_size, /* unmanaged_size */
     NULL, /* describe_refs */
 };
