@@ -3,6 +3,8 @@
 /* This representation's function pointer table. */
 static const MVMREPROps CUnion_this_repr;
 
+static void free_repr_data(MVMCUnionREPRData *repr_data);
+
 /* Locates all of the attributes. Puts them onto a flattened, ordered
  * list of attributes (populating the passed flat_list). Also builds
  * the index mapping for doing named lookups. Note index is not related
@@ -64,6 +66,7 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
             MVM_repr_push_o(tc, attr_map_list, attr_map);
         }
         else {
+            MVM_gc_allocate_gen2_default_clear(tc);
             MVM_exception_throw_adhoc(tc,
                 "CUnion representation does not support multiple inheritance");
         }
@@ -181,6 +184,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(type)->REPR_data;
                         if (!cstruct_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            free_repr_data(repr_data);
                             MVM_exception_throw_adhoc(tc,
                                 "CUnion: can't inline a CStruct attribute before its type's definition");
                         }
@@ -197,6 +202,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCPPStructREPRData *cppstruct_repr_data = (MVMCPPStructREPRData *)STABLE(type)->REPR_data;
                         if (!cppstruct_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            free_repr_data(repr_data);
                             MVM_exception_throw_adhoc(tc,
                                 "CUnion: can't inline a CPPStruct attribute before its type's definition");
                         }
@@ -213,6 +220,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     if (inlined) {
                         MVMCUnionREPRData *cunion_repr_data = (MVMCUnionREPRData *)STABLE(type)->REPR_data;
                         if (!cunion_repr_data) {
+                            MVM_gc_allocate_gen2_default_clear(tc);
+                            free_repr_data(repr_data);
                             MVM_exception_throw_adhoc(tc,
                                 "CUnion: can't inline a CUnion attribute before its type's definition");
                         }
@@ -228,6 +237,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     repr_data->member_types[i] = type;
                 }
                 else {
+                    MVM_gc_allocate_gen2_default_clear(tc);
+                    free_repr_data(repr_data);
                     MVM_exception_throw_adhoc(tc,
                         "CUnion representation only handles attributes of type:\n"
                         "  (u)int8, (u)int16, (u)int32, (u)int64, (u)long, (u)longlong, num32, num64, (s)size_t, bool, Str\n"
@@ -235,12 +246,16 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                 }
             }
             else {
+                MVM_gc_allocate_gen2_default_clear(tc);
+                free_repr_data(repr_data);
                 MVM_exception_throw_adhoc(tc,
                     "CUnion representation requires the types of all attributes to be specified");
             }
 
             if (bits % 8) {
-                 MVM_exception_throw_adhoc(tc,
+                MVM_gc_allocate_gen2_default_clear(tc);
+                free_repr_data(repr_data);
+                MVM_exception_throw_adhoc(tc,
                     "CUnion only supports native types that are a multiple of 8 bits wide (was passed: %"PRId32")", bits);
             }
 
@@ -330,8 +345,8 @@ static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
 /* Composes the representation. */
 static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *repr_info) {
     /* Compute allocation strategy. */
-    MVMCUnionREPRData *repr_data = MVM_calloc(1, sizeof(MVMCUnionREPRData));
     MVMObject *attr_info = MVM_repr_at_key_o(tc, repr_info, tc->instance->str_consts.attribute);
+    MVMCUnionREPRData *repr_data = MVM_calloc(1, sizeof(MVMCUnionREPRData));
     MVM_gc_allocate_gen2_default_set(tc);
     compute_allocation_strategy(tc, attr_info, repr_data);
     MVM_gc_allocate_gen2_default_clear(tc);
@@ -364,9 +379,6 @@ static void initialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, voi
 
 /* Copies to the body of one object to another. */
 static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *dest_root, void *dest) {
-    MVMCUnionREPRData * repr_data = (MVMCUnionREPRData *) st->REPR_data;
-    MVMCUnionBody *src_body = (MVMCUnionBody *)src;
-    MVMCUnionBody *dest_body = (MVMCUnionBody *)dest;
     MVM_exception_throw_adhoc(tc, "cloning a CUnion is NYI");
 }
 
@@ -413,51 +425,52 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                 MVMObject *obj     = body->child_objs[real_slot];
                 if (!obj) {
                     /* No cached object. */
-                    if (repr_data->attribute_locations[slot] & MVM_CUNION_ATTR_INLINED) {
-                        if (type == MVM_CUNION_ATTR_CSTRUCT) {
-                            obj = MVM_nativecall_make_cstruct(tc, typeobj,
-                                (char *)body->cunion + repr_data->struct_offsets[slot]);
-                        }
-                        else if (type == MVM_CUNION_ATTR_CPPSTRUCT) {
-                            obj = MVM_nativecall_make_cppstruct(tc, typeobj,
-                                (char *)body->cunion + repr_data->struct_offsets[slot]);
-                        }
-                        else if (type == MVM_CUNION_ATTR_CUNION) {
-                            obj = MVM_nativecall_make_cunion(tc, typeobj,
-                                (char *)body->cunion + repr_data->struct_offsets[slot]);
-                        }
-                    }
-                    else {
-                        void *cobj = get_ptr_at_offset(body->cunion, repr_data->struct_offsets[slot]);
-                        if (cobj) {
-                            MVMObject **child_objs = body->child_objs;
-                            if (type == MVM_CUNION_ATTR_CARRAY) {
-                                obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
+                    MVMROOT(tc, root, {
+                        if (repr_data->attribute_locations[slot] & MVM_CUNION_ATTR_INLINED) {
+                            if (type == MVM_CUNION_ATTR_CSTRUCT) {
+                                obj = MVM_nativecall_make_cstruct(tc, typeobj,
+                                    (char *)body->cunion + repr_data->struct_offsets[slot]);
                             }
-                            else if(type == MVM_CUNION_ATTR_CSTRUCT) {
-                                obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
+                            else if (type == MVM_CUNION_ATTR_CPPSTRUCT) {
+                                obj = MVM_nativecall_make_cppstruct(tc, typeobj,
+                                    (char *)body->cunion + repr_data->struct_offsets[slot]);
                             }
-                            else if(type == MVM_CUNION_ATTR_CPPSTRUCT) {
-                                obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
-                            }
-                            else if(type == MVM_CUNION_ATTR_CUNION) {
-                                obj = MVM_nativecall_make_cunion(tc, typeobj, cobj);
-                            }
-                            else if(type == MVM_CUNION_ATTR_CPTR) {
-                                obj = MVM_nativecall_make_cpointer(tc, typeobj, cobj);
-                            }
-                            else if(type == MVM_CUNION_ATTR_STRING) {
-                                MVMROOT(tc, typeobj, {
-                                    MVMString *str = MVM_string_utf8_decode(tc, tc->instance->VMString,
-                                        cobj, strlen(cobj));
-                                    obj = MVM_repr_box_str(tc, typeobj, str);
-                                });
+                            else if (type == MVM_CUNION_ATTR_CUNION) {
+                                obj = MVM_nativecall_make_cunion(tc, typeobj,
+                                    (char *)body->cunion + repr_data->struct_offsets[slot]);
                             }
                         }
                         else {
-                            obj = typeobj;
+                            void *cobj = get_ptr_at_offset(body->cunion, repr_data->struct_offsets[slot]);
+                            if (cobj) {
+                                if (type == MVM_CUNION_ATTR_CARRAY) {
+                                    obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
+                                }
+                                else if(type == MVM_CUNION_ATTR_CSTRUCT) {
+                                    obj = MVM_nativecall_make_cstruct(tc, typeobj, cobj);
+                                }
+                                else if(type == MVM_CUNION_ATTR_CPPSTRUCT) {
+                                    obj = MVM_nativecall_make_cppstruct(tc, typeobj, cobj);
+                                }
+                                else if(type == MVM_CUNION_ATTR_CUNION) {
+                                    obj = MVM_nativecall_make_cunion(tc, typeobj, cobj);
+                                }
+                                else if(type == MVM_CUNION_ATTR_CPTR) {
+                                    obj = MVM_nativecall_make_cpointer(tc, typeobj, cobj);
+                                }
+                                else if(type == MVM_CUNION_ATTR_STRING) {
+                                    MVMROOT(tc, typeobj, {
+                                        MVMString *str = MVM_string_utf8_decode(tc, tc->instance->VMString,
+                                            cobj, strlen(cobj));
+                                        obj = MVM_repr_box_str(tc, typeobj, str);
+                                    });
+                                }
+                            }
+                            else {
+                                obj = typeobj;
+                            }
                         }
-                    }
+                    });
                     MVM_ASSIGN_REF(tc, &(root->header), body->child_objs[real_slot], obj);
                 }
                 result_reg->o = obj;
@@ -643,9 +656,27 @@ static void gc_mark_repr_data(MVMThreadContext *tc, MVMSTable *st, MVMGCWorklist
     }
 }
 
+static void free_repr_data(MVMCUnionREPRData *repr_data) {
+    /* May not have survived to composition. */
+    if (repr_data == NULL)
+        return;
+
+    if (repr_data->name_to_index_mapping) {
+        MVM_free(repr_data->name_to_index_mapping);
+        MVM_free(repr_data->attribute_locations);
+        MVM_free(repr_data->struct_offsets);
+        MVM_free(repr_data->flattened_stables);
+        MVM_free(repr_data->member_types);
+        MVM_free(repr_data->initialize_slots);
+    }
+
+    MVM_free(repr_data);
+}
+
 /* Free representation data. */
 static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
-    MVM_free(st->REPR_data);
+    MVMCUnionREPRData *repr_data = (MVMCUnionREPRData *)st->REPR_data;
+    free_repr_data(repr_data);
 }
 
 /* This is called to do any cleanup of resources when an object gets
