@@ -104,8 +104,8 @@ static const MVMREPROps MVMMultiCache_this_repr = {
 
 /* Filters for various parts of action.arg_match. */
 #define MVM_MULTICACHE_ARG_IDX_FILTER  (2 * MVM_INTERN_ARITY_LIMIT - 1)
-#define MVM_MULTICACHE_ARG_CONC_FILTER 0x10
-#define MVM_MULTICACHE_ARG_RW_FILTER   0x20
+#define MVM_MULTICACHE_ARG_CONC_FILTER (2 * MVM_INTERN_ARITY_LIMIT)
+#define MVM_MULTICACHE_ARG_RW_FILTER   (4 * MVM_INTERN_ARITY_LIMIT)
 #define MVM_MULTICACHE_TYPE_ID_FILTER  (0xFFFFFFFFFFFFFFFFULL ^ (MVM_TYPE_CACHE_ID_INCR - 1))
 
 /* Debug support dumps the tree after each addition. */
@@ -114,7 +114,7 @@ static const MVMREPROps MVMMultiCache_this_repr = {
 static void dump_cache(MVMThreadContext *tc, MVMMultiCacheBody *cache) {
     MVMint32 num_nodes = cache->cache_memory_size / sizeof(MVMMultiCacheNode);
     MVMint32 i;
-    printf("Multi cache at %p (%d nodes, %d results)\n",
+    printf("Multi cache at %p (%d nodes, %zd results)\n",
         cache, num_nodes, cache->num_results);
     for (i = 0; i < num_nodes; i++)
         printf(" - %p -> (Y: %d, N: %d)\n",
@@ -204,13 +204,10 @@ MVMObject * MVM_multi_cache_add(MVMThreadContext *tc, MVMObject *cache_obj, MVMO
         }
     }
 
-    /* Oobtain the cache addition lock, and then do another lookup to ensure
-     * nobody beat us to making this entry. */
+    /* Obtain the cache addition lock. */
     uv_mutex_lock(&(tc->instance->mutex_multi_cache_add));
-    if (MVM_multi_cache_find(tc, cache_obj, capture))
-        goto DONE;
 
-    /* We're now udner the insertion lock and know nobody else can tweak the
+    /* We're now under the insertion lock and know nobody else can tweak the
      * cache. First, see if there's even a current version and search tree. */
     have_head = 0;
     have_tree = 0;
@@ -238,25 +235,46 @@ MVMObject * MVM_multi_cache_add(MVMThreadContext *tc, MVMObject *cache_obj, MVMO
 
         /* Chase until we reach an arg we don't match. */
         while (cur_node > 0) {
-            MVMuint64 arg_match = tree[cur_node].action.arg_match;
-            MVMuint64 arg_idx   = arg_match & MVM_MULTICACHE_ARG_IDX_FILTER;
+            MVMuint64    arg_match = tree[cur_node].action.arg_match;
+            MVMuint64    arg_idx   = arg_match & MVM_MULTICACHE_ARG_IDX_FILTER;
+            MVMuint64    type_id   = arg_match & MVM_MULTICACHE_TYPE_ID_FILTER;
+            MVMRegister  arg       = apc->args[arg_idx];
+            MVMSTable   *st        = STABLE(arg.o);
+            MVMuint64    is_rw     = 0;
             tweak_node = cur_node;
-            if ((match_flags[arg_idx] | arg_idx) == arg_match) {
-                matched_args++;
-                unmatched_arg = 0;
-                cur_node = tree[cur_node].match;
+            if (st->container_spec && IS_CONCRETE(arg.o)) {
+                MVMContainerSpec const *contspec = st->container_spec;
+                if (!contspec->fetch_never_invokes)
+                    goto DONE;
+                if (REPR(arg.o)->ID != MVM_REPR_ID_NativeRef) {
+                    is_rw = contspec->can_store(tc, arg.o);
+                    contspec->fetch(tc, arg.o, &arg);
+                }
+                else {
+                    is_rw = 1;
+                }
             }
-            else {
-                unmatched_arg = 1;
-                cur_node = tree[cur_node].no_match;
+            if (STABLE(arg.o)->type_cache_id == type_id) {
+                MVMuint32 need_concrete = (arg_match & MVM_MULTICACHE_ARG_CONC_FILTER) ? 1 : 0;
+                if (IS_CONCRETE(arg.o) == need_concrete) {
+                    MVMuint32 need_rw = (arg_match & MVM_MULTICACHE_ARG_RW_FILTER) ? 1 : 0;
+                    if (need_rw == is_rw) {
+                        matched_args++;
+                        unmatched_arg = 0;
+                        cur_node = tree[cur_node].match;
+                        continue;
+                    }
+                }
             }
+            unmatched_arg = 1;
+            cur_node = tree[cur_node].no_match;
         }
 
-        /* If we found a candidate, something inconsistent, as we
-         * checked for non-entry above. */
+        /* If we found a candidate, someone else beat us to adding the candidate
+           before we obtained the lock or the arguments got changed behind our
+           back so that they now match */
         if (cur_node != 0)
-            MVM_panic(1, "Corrupt multi dispatch cache: cur_node != 0, re-check == %p",
-                MVM_multi_cache_find(tc, cache_obj, capture));
+            goto DONE;
     }
 
     /* Now calculate the new size we'll need to allocate. */
@@ -487,8 +505,9 @@ MVMObject * MVM_multi_cache_find_spesh(MVMThreadContext *tc, MVMObject *cache_ob
             ? arg_info->arg_facts[arg_idx]
             : NULL;
         if (type_tuple) {
-            MVMuint64 tt_offset = arg_idx >= arg_info->cs->num_pos
-                ? ((arg_idx + 1) - arg_info->cs->num_pos) / 2
+            MVMuint16 num_pos = arg_info->cs->num_pos;
+            MVMuint64 tt_offset = arg_idx >= num_pos
+                ? ((arg_idx - 1) - num_pos) / 2 + num_pos
                 : arg_idx;
             MVMuint32 is_rw = type_tuple[tt_offset].rw_cont;
             MVMSTable *known_type_st = NULL;
