@@ -461,13 +461,9 @@ void MVM_args_set_result_obj(MVMThreadContext *tc, MVMObject *result, MVMint32 f
     else {
         MVMROOT(tc, result, {
             if (MVM_spesh_log_is_caller_logging(tc))
-                MVMROOT(tc, result, {
-                    MVM_spesh_log_return_type(tc, result);
-                });
+                MVM_spesh_log_return_type(tc, result);
             else if (MVM_spesh_log_is_logging(tc))
-                MVMROOT(tc, result, {
-                    MVM_spesh_log_return_to_unlogged(tc);
-                });
+                MVM_spesh_log_return_to_unlogged(tc);
         });
         target = tc->cur_frame->caller;
     }
