@@ -26,6 +26,9 @@ MVMint64 MVM_coerce_istrue_s(MVMThreadContext *tc, MVMString *str) {
  * next instruction before this is called. */
 static void boolify_return(MVMThreadContext *tc, void *sr_data);
 static void flip_return(MVMThreadContext *tc, void *sr_data);
+static void free_boolify_return_data(MVMThreadContext *tc, void *sr_data) {
+    MVM_free(sr_data);
+}
 void MVM_coerce_istrue(MVMThreadContext *tc, MVMObject *obj, MVMRegister *res_reg,
         MVMuint8 *true_addr, MVMuint8 *false_addr, MVMuint8 flip) {
     MVMint64 result = 0;
@@ -53,7 +56,7 @@ void MVM_coerce_istrue(MVMThreadContext *tc, MVMObject *obj, MVMRegister *res_re
                     data->true_addr  = true_addr;
                     data->false_addr = false_addr;
                     data->flip       = flip;
-                    MVM_frame_special_return(tc, tc->cur_frame, boolify_return, NULL, data, NULL);
+                    MVM_frame_special_return(tc, tc->cur_frame, boolify_return, free_boolify_return_data, data, NULL);
                     MVM_args_setup_thunk(tc, &data->res_reg, MVM_RETURN_INT, inv_arg_callsite);
                     tc->cur_frame->args[0].o = obj;
                     STABLE(code)->invoke(tc, code, inv_arg_callsite, tc->cur_frame->args);
@@ -184,7 +187,6 @@ static char * u64toa_naive_worker(uint64_t value, char* buffer) {
         *buffer++ = *--p;
     } while (p != temp);
 
-    *buffer = '\0';
     return buffer;
 }
 static size_t i64toa_naive(int64_t value, char* buffer) {
@@ -201,7 +203,7 @@ static size_t u64toa_naive(uint64_t value, char* buffer) {
     return u64toa_naive_worker(value, buffer) - buffer;
 }
 MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
-    char buffer[64];
+    char buffer[20];
     int len;
     /* See if we can hit the cache. */
     int cache = 0 <= i && i < MVM_INT_TO_STR_CACHE_SIZE;
@@ -227,7 +229,7 @@ MVMString * MVM_coerce_i_s(MVMThreadContext *tc, MVMint64 i) {
 }
 
 MVMString * MVM_coerce_u_s(MVMThreadContext *tc, MVMuint64 i) {
-    char buffer[64];
+    char buffer[20];
     int len;
     /* See if we can hit the cache. */
     int cache = i < MVM_INT_TO_STR_CACHE_SIZE;
@@ -326,7 +328,7 @@ void MVM_coerce_smart_stringify(MVMThreadContext *tc, MVMObject *obj, MVMRegiste
 }
 
 MVMint64 MVM_coerce_s_i(MVMThreadContext *tc, MVMString *s) {
-    char     *enc = MVM_string_ascii_encode(tc, s, NULL, 0);
+    char     *enc = MVM_string_ascii_encode_any(tc, s);
     MVMint64  i   = strtoll(enc, NULL, 10);
     MVM_free(enc);
     return i;
@@ -451,9 +453,6 @@ MVMint64 MVM_coerce_simple_intify(MVMThreadContext *tc, MVMObject *obj) {
     }
 }
 
-/* concatenating with "" ensures that only literal strings are accepted as argument. */
-#define STR_WITH_LEN(str)  ("" str ""), (sizeof(str) - 1)
-
 MVMObject * MVM_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *str, MVMint64 offset, MVMint64 flag) {
     MVMObject *result;
     MVMint64 zvalue = 0;
@@ -526,58 +525,3 @@ MVMObject * MVM_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *str, MVMi
 
     return result;
 }
-
-
-void MVM_box_int(MVMThreadContext *tc, MVMint64 value, MVMObject *type,
-             MVMRegister * dst) {
-    MVMObject *box = MVM_intcache_get(tc, type, value);
-    if (box == 0) {
-        box = REPR(type)->allocate(tc, STABLE(type));
-        if (REPR(box)->initialize)
-            REPR(box)->initialize(tc, STABLE(box), box, OBJECT_BODY(box));
-        REPR(box)->box_funcs.set_int(tc, STABLE(box), box,
-                                     OBJECT_BODY(box), value);
-    }
-    dst->o = box;
-}
-
-void MVM_box_num(MVMThreadContext *tc, MVMnum64 value, MVMObject *type,
-                 MVMRegister * dst) {
-    MVMObject *box = REPR(type)->allocate(tc, STABLE(type));
-    if (REPR(box)->initialize)
-        REPR(box)->initialize(tc, STABLE(box), box, OBJECT_BODY(box));
-    REPR(box)->box_funcs.set_num(tc, STABLE(box), box,
-                                 OBJECT_BODY(box), value);
-    dst->o = box;
-
-}
-
-MVMString * MVM_unbox_str(MVMThreadContext *tc, MVMObject *obj) {
-    if (!IS_CONCRETE(obj))
-        MVM_exception_throw_adhoc(tc, "Cannot unbox a type object (%s) to a str.",
-            MVM_6model_get_debug_name(tc, obj));
-    return REPR(obj)->box_funcs.get_str(tc,
-        STABLE(obj), obj, OBJECT_BODY(obj));
-}
-
-void MVM_box_str(MVMThreadContext *tc, MVMString *value, MVMObject *type,
-                 MVMRegister *dst) {
-    MVMObject *box;
-    MVMROOT(tc, value, {
-        box = REPR(type)->allocate(tc, STABLE(type));
-        if (REPR(box)->initialize)
-            REPR(box)->initialize(tc, STABLE(box), box, OBJECT_BODY(box));
-        REPR(box)->box_funcs.set_str(tc, STABLE(box), box,
-                                     OBJECT_BODY(box), value);
-        dst->o = box;
-    });
-}
-
-void MVM_box_uint(MVMThreadContext *tc, MVMuint64 value, MVMObject *type,
-             MVMRegister * dst) {
-    MVMObject *box = REPR(type)->allocate(tc, STABLE(type));
-    if (REPR(box)->initialize)
-        REPR(box)->initialize(tc, STABLE(box), box, OBJECT_BODY(box));
-    REPR(box)->box_funcs.set_uint(tc, STABLE(box), box, OBJECT_BODY(box), value);
-    dst->o = box;
-}
