@@ -359,7 +359,7 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
 
     case MVM_EX_ACTION_INVOKE: {
         /* Create active handler record. */
-        MVMActiveHandler *ah = MVM_malloc(sizeof(MVMActiveHandler));
+        MVMActiveHandler *ah = MVM_fixed_size_alloc(tc, tc->instance->fsa, sizeof(MVMActiveHandler));
         MVMFrame *cur_frame = tc->cur_frame;
         MVMFrame *pres_frame;
         MVMObject *handler_code;
@@ -440,7 +440,7 @@ static void unwind_after_handler(MVMThreadContext *tc, void *sr_data) {
     }
     /* Clean up. */
     tc->active_handlers = ah->next_handler;
-    MVM_free(ah);
+    MVM_fixed_size_free(tc, tc->instance->fsa, sizeof(MVMActiveHandler), ah);
 
     /* Do the unwinding as needed. */
     if (exception && exception->body.return_after_unwind) {
@@ -462,7 +462,7 @@ static void cleanup_active_handler(MVMThreadContext *tc, void *sr_data) {
 
     /* Clean up. */
     tc->active_handlers = ah->next_handler;
-    MVM_free(ah);
+    MVM_fixed_size_free(tc, tc->instance->fsa, sizeof(MVMActiveHandler), ah);
 }
 
 char * MVM_exception_backtrace_line(MVMThreadContext *tc, MVMFrame *cur_frame,
@@ -514,6 +514,7 @@ char * MVM_exception_backtrace_line(MVMThreadContext *tc, MVMFrame *cur_frame,
 /* Returns a list of hashes containing file, line, sub and annotations. */
 MVMObject * MVM_exception_backtrace(MVMThreadContext *tc, MVMObject *ex_obj) {
     MVMFrame *cur_frame;
+    MVMSpeshFrameWalker fw;
     MVMObject *arr = NULL, *annotations = NULL, *row = NULL, *value = NULL;
     MVMuint32 count = 0;
     MVMString *k_file = NULL, *k_line = NULL, *k_sub = NULL, *k_anno = NULL;
@@ -523,6 +524,10 @@ MVMObject * MVM_exception_backtrace(MVMThreadContext *tc, MVMObject *ex_obj) {
         cur_frame = ((MVMException *)ex_obj)->body.origin;
         throw_address = ((MVMException *)ex_obj)->body.throw_address;
     }
+    else if (ex_obj == tc->instance->VMNull) {
+        cur_frame = tc->cur_frame;
+        throw_address = *tc->interp_cur_op;
+    }
     else {
         MVM_exception_throw_adhoc(tc, "Op 'backtrace' needs an exception object");
     }
@@ -544,46 +549,50 @@ MVMObject * MVM_exception_backtrace(MVMThreadContext *tc, MVMObject *ex_obj) {
 
     arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
 
-    while (cur_frame != NULL) {
-        MVMuint8             *cur_op = count ? cur_frame->return_address : throw_address;
-        MVMuint32             offset = cur_op - MVM_frame_effective_bytecode(cur_frame);
-        MVMBytecodeAnnotation *annot = MVM_bytecode_resolve_annotation(tc, &cur_frame->static_info->body,
-                                            offset > 0 ? offset - 1 : 0);
-        MVMuint32             fshi   = annot ? (MVMint32)annot->filename_string_heap_index : -1;
-        char            *line_number = MVM_malloc(16);
-        MVMString      *filename_str;
-        snprintf(line_number, 16, "%d", annot ? annot->line_number : 1);
+    if (cur_frame) {
+        MVM_spesh_frame_walker_init(tc, &fw, cur_frame, 0);
+        MVM_spesh_frame_walker_next(tc, &fw); // Go to first inline
 
-        /* annotations hash will contain "file" and "line" */
-        annotations = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTHash);
+        do {
+            cur_frame = MVM_spesh_frame_walker_current_frame(tc, &fw);
+            MVMuint8             *cur_op = count ? cur_frame->return_address : throw_address;
+            MVMuint32             offset = cur_op - MVM_frame_effective_bytecode(cur_frame);
+            MVMBytecodeAnnotation *annot = MVM_bytecode_resolve_annotation(tc, &cur_frame->static_info->body,
+                                                offset > 0 ? offset - 1 : 0);
+            MVMuint32             fshi   = annot ? (MVMint32)annot->filename_string_heap_index : -1;
+            char         line_number[11] = {0}; // 11 == 10 (max # of decimal digits in an MVMuint32) + 1 (NULL terminator)
+            MVMString      *filename_str;
+            snprintf(line_number, 11, "%d", annot ? annot->line_number : 1);
 
-        /* file */
-        filename_str = annot && fshi < cur_frame->static_info->body.cu->body.num_strings
-             ? MVM_cu_string(tc, cur_frame->static_info->body.cu, fshi)
-             : cur_frame->static_info->body.cu->body.filename;
-        value = MVM_repr_box_str(tc, MVM_hll_current(tc)->str_box_type,
-            filename_str ? filename_str : tc->instance->str_consts.empty);
-        MVM_repr_bind_key_o(tc, annotations, k_file, value);
+            /* annotations hash will contain "file" and "line" */
+            annotations = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTHash);
 
-        /* line */
-        value = (MVMObject *)MVM_string_ascii_decode_nt(tc, tc->instance->VMString, line_number);
-        value = MVM_repr_box_str(tc, MVM_hll_current(tc)->str_box_type, (MVMString *)value);
-        MVM_repr_bind_key_o(tc, annotations, k_line, value);
-        MVM_free(line_number);
+            /* file */
+            filename_str = annot && fshi < cur_frame->static_info->body.cu->body.num_strings
+                 ? MVM_cu_string(tc, cur_frame->static_info->body.cu, fshi)
+                 : cur_frame->static_info->body.cu->body.filename;
+            value = MVM_repr_box_str(tc, MVM_hll_current(tc)->str_box_type,
+                filename_str ? filename_str : tc->instance->str_consts.empty);
+            MVM_repr_bind_key_o(tc, annotations, k_file, value);
 
-        /* row will contain "sub" and "annotations" */
-        row = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTHash);
-        MVM_repr_bind_key_o(tc, row, k_sub,
-            cur_frame->code_ref ? cur_frame->code_ref : tc->instance->VMNull);
-        MVM_repr_bind_key_o(tc, row, k_anno, annotations);
+            /* line */
+            value = (MVMObject *)MVM_string_ascii_decode_nt(tc, tc->instance->VMString, line_number);
+            value = MVM_repr_box_str(tc, MVM_hll_current(tc)->str_box_type, (MVMString *)value);
+            MVM_repr_bind_key_o(tc, annotations, k_line, value);
 
-        MVM_repr_push_o(tc, arr, row);
-        MVM_free(annot);
+            /* row will contain "sub" and "annotations" */
+            row = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTHash);
+            MVM_repr_bind_key_o(tc, row, k_sub,
+                cur_frame->code_ref ? cur_frame->code_ref : tc->instance->VMNull);
+            MVM_repr_bind_key_o(tc, row, k_anno, annotations);
 
-        cur_frame = cur_frame->caller;
-        while (cur_frame && cur_frame->static_info->body.is_thunk)
-            cur_frame = cur_frame->caller;
-        count++;
+            MVM_repr_push_o(tc, arr, row);
+            MVM_free(annot);
+
+            count++;
+        } while (MVM_spesh_frame_walker_move_caller_skip_thunks(tc, &fw));
+
+        MVM_spesh_frame_walker_cleanup(tc, &fw);
     }
 
     MVM_gc_root_temp_pop_n(tc, 9);
@@ -829,7 +838,7 @@ void MVM_exception_resume(MVMThreadContext *tc, MVMObject *ex_obj) {
     /* Clear the current active handler. */
     ah = tc->active_handlers;
     tc->active_handlers = ah->next_handler;
-    MVM_free(ah);
+    MVM_fixed_size_free(tc, tc->instance->fsa, sizeof(MVMActiveHandler), ah);
 
     /* Unwind to the thrower of the exception; set PC and jit entry label. */
     MVM_frame_unwind_to(tc, target, ex->body.resume_addr, 0, NULL, ex->body.jit_resume_label);
