@@ -187,26 +187,24 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
     add_collectable(tc, worklist, snapshot, tc->next_dispatcher_for, "Next dispatcher for");
 
     /* Callback cache. */
-    if (tc->native_callback_cache) {
-        MVMStrHashTable *cache = tc->native_callback_cache;
-        MVMStrHashIterator iterator = MVM_str_hash_first(tc, cache);
-        while (!MVM_str_hash_at_end(tc, cache, iterator)) {
-            struct MVMNativeCallbackCacheHead *current_cbceh
-                = MVM_str_hash_current_nocheck(tc, cache, iterator);
-            MVMint32 i;
-            MVMNativeCallback *entry = current_cbceh->head;
-            add_collectable(tc, worklist, snapshot, current_cbceh->hash_handle.key,
-                            "Native callback cache key");
-            while (entry) {
-                for (i = 0; i < entry->num_types; i++)
-                    add_collectable(tc, worklist, snapshot, entry->types[i],
-                                    "Native callback cache type");
-                add_collectable(tc, worklist, snapshot, entry->target,
-                                "Native callback cache target");
-                entry = entry->next;
-            }
-            iterator = MVM_str_hash_next_nocheck(tc, cache, iterator);
+    MVMStrHashTable *cache = &tc->native_callback_cache;
+    MVMStrHashIterator iterator = MVM_str_hash_first(tc, cache);
+    while (!MVM_str_hash_at_end(tc, cache, iterator)) {
+        struct MVMNativeCallbackCacheHead *current_cbceh
+            = MVM_str_hash_current_nocheck(tc, cache, iterator);
+        MVMint32 i;
+        MVMNativeCallback *entry = current_cbceh->head;
+        add_collectable(tc, worklist, snapshot, current_cbceh->hash_handle.key,
+                        "Native callback cache key");
+        while (entry) {
+            for (i = 0; i < entry->num_types; i++)
+                add_collectable(tc, worklist, snapshot, entry->types[i],
+                                "Native callback cache type");
+            add_collectable(tc, worklist, snapshot, entry->target,
+                            "Native callback cache target");
+            entry = entry->next;
         }
+        iterator = MVM_str_hash_next_nocheck(tc, cache, iterator);
     }
 
     /* Profiling data. */
@@ -416,6 +414,7 @@ void MVM_gc_root_add_frame_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist
     MVM_gc_worklist_add(tc, worklist, &cur_frame->outer);
     MVM_gc_worklist_add(tc, worklist, &cur_frame->code_ref);
     MVM_gc_worklist_add(tc, worklist, &cur_frame->static_info);
+    MVM_gc_worklist_add(tc, worklist, &cur_frame->spesh_cand);
 
     /* Mark frame extras if needed. */
     if (cur_frame->extra) {
@@ -450,13 +449,13 @@ void MVM_gc_root_add_frame_registers_to_worklist(MVMThreadContext *tc, MVMGCWork
         /* Scan locals. */
 
         MVMSpeshCandidate *spesh_cand = frame->spesh_cand;
-        MVMJitCode *jitcode = spesh_cand ? spesh_cand->jitcode : NULL;
+        MVMJitCode *jitcode = spesh_cand ? spesh_cand->body.jitcode : NULL;
         if (jitcode && jitcode->local_types) {
             type_map = jitcode->local_types;
             count    = jitcode->num_locals;
-        } else if (frame->spesh_cand && frame->spesh_cand->local_types) {
-            type_map = frame->spesh_cand->local_types;
-            count    = frame->spesh_cand->num_locals;
+        } else if (spesh_cand && spesh_cand->body.local_types) {
+            type_map = spesh_cand->body.local_types;
+            count    = spesh_cand->body.num_locals;
         }
         else {
             type_map = frame->static_info->body.local_types;
@@ -505,9 +504,9 @@ static void scan_lexicals(MVMThreadContext *tc, MVMGCWorklist *worklist, MVMFram
     if (frame->env) {
         MVMuint16  i, count;
         MVMuint16 *type_map;
-        if (frame->spesh_cand && frame->spesh_cand->lexical_types) {
-            type_map = frame->spesh_cand->lexical_types;
-            count    = frame->spesh_cand->num_lexicals;
+        if (frame->spesh_cand && frame->spesh_cand->body.lexical_types) {
+            type_map = frame->spesh_cand->body.lexical_types;
+            count    = frame->spesh_cand->body.num_lexicals;
         }
         else {
             type_map = frame->static_info->body.lexical_types;
