@@ -44,15 +44,11 @@ static int mkdir_p(MVMThreadContext *tc, char *pathname, MVMint64 mode) {
             ch = *p;
             *p  = '\0';
 #ifdef _WIN32
-            if (CreateDirectoryW(pathname, NULL)) {
-                created = 1;
-            }
+            created = CreateDirectoryW(pathname, NULL);
 #else
-            if (uv_fs_stat(NULL, &req, pathname, NULL) <= 0) {
-                if (mkdir(pathname, mode) != -1) {
-                    created = 1;
-                }
-            }
+            created = (uv_fs_stat(NULL, &req, pathname, NULL) == 0
+                       && S_ISDIR(req.statbuf.st_mode)
+                       || mkdir(pathname, mode) == 0);
 #endif
             if (!(*p = ch)) break;
         }
@@ -102,7 +98,7 @@ void MVM_dir_mkdir(MVMThreadContext *tc, MVMString *path, MVMint64 mode) {
     MVM_free(wpathname);
 #else
 
-    if (mkdir_p(tc, pathname, mode) == -1 && errno != EEXIST) {
+    if (mkdir_p(tc, pathname, mode) == -1) {
         int mkdir_error = errno;
         MVM_free(pathname);
         MVM_exception_throw_adhoc(tc, "Failed to mkdir: %s", strerror(mkdir_error));
@@ -204,60 +200,69 @@ static const MVMIOOps op_table = {
 
 /* Open a filehandle, returning a handle. */
 MVMObject * MVM_dir_open(MVMThreadContext *tc, MVMString *dirname) {
-    MVMOSHandle  * const result = (MVMOSHandle *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIO);
+    MVMOSHandle  * result;
     MVMIODirIter * const data   = MVM_calloc(1, sizeof(MVMIODirIter));
+    MVMROOT(tc, dirname, {
+        result = (MVMOSHandle *)MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTIO);
+    });
 #ifdef _WIN32
-    char *name;
-    int str_len;
-    wchar_t *wname;
-    wchar_t *dir_name;
+    {
+        char *name;
+        int str_len;
+        wchar_t *wname;
+        wchar_t *dir_name;
 
-    name  = MVM_string_utf8_c8_encode_C_string(tc, dirname);
-    wname = UTF8ToUnicode(name);
-    MVM_free(name);
+        name  = MVM_string_utf8_c8_encode_C_string(tc, dirname);
+        wname = UTF8ToUnicode(name);
+        MVM_free(name);
 
-    str_len = wcslen(wname);
+        str_len = wcslen(wname);
 
-    if (str_len > MAX_PATH - 2) { // the length of later appended '\*' is 2
-        wchar_t  abs_dirname[4096]; /* 4096 should be enough for absolute path */
-        wchar_t *lpp_part;
+        if (str_len > MAX_PATH - 2) { // the length of later appended '\*' is 2
+            wchar_t  abs_dirname[4096]; /* 4096 should be enough for absolute path */
+            wchar_t *lpp_part;
 
-        /* You cannot use the "\\?\" prefix with a relative path,
-         * relative paths are always limited to a total of MAX_PATH characters.
-         * see http://msdn.microsoft.com/en-us/library/windows/desktop/aa365247%28v=vs.85%29.aspx */
-        if (!GetFullPathNameW(wname, 4096, abs_dirname, &lpp_part)) {
+            /* You cannot use the "\\?\" prefix with a relative path,
+             * relative paths are always limited to a total of MAX_PATH characters.
+             * see http://msdn.microsoft.com/en-us/library/windows/desktop/aa365247%28v=vs.85%29.aspx */
+            if (!GetFullPathNameW(wname, 4096, abs_dirname, &lpp_part)) {
+                MVM_free(data);
+                MVM_free(wname);
+                MVM_exception_throw_adhoc(tc, "Directory path is wrong: %d", GetLastError());
+            }
+            MVM_free(wname);
+
+            str_len  = wcslen(abs_dirname);
+            dir_name = (wchar_t *)MVM_malloc((str_len + 7) * sizeof(wchar_t));
+            wcscpy(dir_name, L"\\\\?\\");
+            wcscat(dir_name, abs_dirname);
+        } else {
+            dir_name = (wchar_t *)MVM_malloc((str_len + 3) * sizeof(wchar_t));
+            wcscpy(dir_name, wname);
             MVM_free(wname);
-            MVM_exception_throw_adhoc(tc, "Directory path is wrong: %d", GetLastError());
         }
-        MVM_free(wname);
 
-        str_len  = wcslen(abs_dirname);
-        dir_name = (wchar_t *)MVM_malloc((str_len + 7) * sizeof(wchar_t));
-        wcscpy(dir_name, L"\\\\?\\");
-        wcscat(dir_name, abs_dirname);
-    } else {
-        dir_name = (wchar_t *)MVM_malloc((str_len + 3) * sizeof(wchar_t));
-        wcscpy(dir_name, wname);
-        MVM_free(wname);
+        wcscat(dir_name, L"\\*");     /* Three characters are for the "\*" plus NULL appended.
+                                       * see http://msdn.microsoft.com/en-us/library/windows/desktop/aa365200%28v=vs.85%29.aspx */
+
+        data->dir_name   = dir_name;
+        data->dir_handle = INVALID_HANDLE_VALUE;
     }
 
-    wcscat(dir_name, L"\\*");     /* Three characters are for the "\*" plus NULL appended.
-                                   * see http://msdn.microsoft.com/en-us/library/windows/desktop/aa365200%28v=vs.85%29.aspx */
-
-    data->dir_name   = dir_name;
-    data->dir_handle = INVALID_HANDLE_VALUE;
-
 #else
-    char * const dir_name = MVM_string_utf8_c8_encode_C_string(tc, dirname);
-    DIR * const dir_handle = opendir(dir_name);
-    int opendir_error = errno;
-    MVM_free(dir_name);
+    {
+        char * const dir_name = MVM_string_utf8_c8_encode_C_string(tc, dirname);
+        DIR * const dir_handle = opendir(dir_name);
+        int opendir_error = errno;
+        MVM_free(dir_name);
 
-    if (!dir_handle) {
-        MVM_exception_throw_adhoc(tc, "Failed to open dir: %s", strerror(opendir_error));
+        if (!dir_handle) {
+            MVM_free(data);
+            MVM_exception_throw_adhoc(tc, "Failed to open dir: %s", strerror(opendir_error));
+        }
+
+        data->dir_handle = dir_handle;
     }
-
-    data->dir_handle = dir_handle;
 #endif
 
     result->body.ops  = &op_table;
@@ -286,7 +291,10 @@ MVMString * MVM_dir_read(MVMThreadContext *tc, MVMObject *oshandle) {
     WIN32_FIND_DATAW ffd;
     char *dir_str;
 
-    if (data->dir_handle == INVALID_HANDLE_VALUE) {
+    if (!data->dir_handle) {
+        MVM_exception_throw_adhoc(tc, "Cannot read a closed dir handle.");
+    }
+    else if (data->dir_handle == INVALID_HANDLE_VALUE) {
         HANDLE hFind = FindFirstFileW(data->dir_name, &ffd);
 
         if (hFind == INVALID_HANDLE_VALUE) {
@@ -337,12 +345,13 @@ void MVM_dir_close(MVMThreadContext *tc, MVMObject *oshandle) {
 
 #ifdef _WIN32
     if (data->dir_name) {
-        MVM_free(data->dir_name);
-        data->dir_name = NULL;
+        MVM_free_null(data->dir_name);
     }
 
-    if (!FindClose(data->dir_handle))
-        MVM_exception_throw_adhoc(tc, "Failed to close dirhandle: %d", GetLastError());
+    if (data->dir_handle != INVALID_HANDLE_VALUE) {
+        if (!FindClose(data->dir_handle))
+            MVM_exception_throw_adhoc(tc, "Failed to close dirhandle: %d", GetLastError());
+    }
     data->dir_handle = NULL;
 #else
     if (closedir(data->dir_handle) == -1)
