@@ -122,7 +122,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     instance->main_thread = MVM_tc_create(NULL, instance);
 
     instance->subscriptions.vm_startup_hrtime = uv_hrtime();
-    instance->subscriptions.vm_startup_now = MVM_proc_time_n(instance->main_thread);
+    instance->subscriptions.vm_startup_now = MVM_proc_time(instance->main_thread);
 
 #if MVM_HASH_RANDOMIZE
     /* Get the 128-bit hashSecret */
@@ -170,7 +170,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     MVM_fixkey_hash_build(instance->main_thread, &instance->compilee_hll_configs, sizeof(MVMHLLConfig));
 
     /* Set up DLL registry mutex. */
-    init_mutex(instance->mutex_dll_registry, "REPR registry");
+    init_mutex(instance->mutex_dll_registry, "DLL registry");
     MVM_fixkey_hash_build(instance->main_thread, &instance->dll_registry, sizeof(struct MVMDLLRegistry));
 
     /* Set up extension registry mutex. */
@@ -590,8 +590,13 @@ void MVM_vm_exit(MVMInstance *instance) {
     MVM_io_flush_standard_handles(instance->main_thread);
 
     /* Close any spesh or jit log. */
-    if (instance->spesh_log_fh)
+    if (instance->spesh_log_fh) {
+        /* Need to properly shut down spesh, otherwise we may segfault trying
+         * to write to a closed file handle before we exit completely */
+        MVM_spesh_worker_stop(instance->main_thread);
+        MVM_spesh_worker_join(instance->main_thread);
         fclose(instance->spesh_log_fh);
+    }
     if (instance->dynvar_log_fh) {
         fprintf(instance->dynvar_log_fh, "- x 0 0 0 0 %"PRId64" %"PRIu64" %"PRIu64"\n", instance->dynvar_log_lasttime, uv_hrtime(), uv_hrtime());
         fclose(instance->dynvar_log_fh);
@@ -627,6 +632,11 @@ static void cleanup_callsite_interns(MVMInstance *instance) {
     MVM_free(instance->callsite_interns);
 }
 
+static void free_lib(MVMThreadContext *tc, void *entry_v, void *arg) {
+    struct MVMDLLRegistry *entry = entry_v;
+    MVM_nativecall_free_lib(entry->lib);
+}
+
 /* Destroys a VM instance. This must be called only from the main thread. It
  * should clear up all resources and free all memory; in practice, it falls
  * short of this goal at the moment. */
@@ -680,6 +690,7 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
 
     /* Clean up Hash of DLLs. */
     uv_mutex_destroy(&instance->mutex_dll_registry);
+    MVM_fixkey_hash_foreach(instance->main_thread, &instance->dll_registry, free_lib, NULL);
     MVM_fixkey_hash_demolish(instance->main_thread, &instance->dll_registry);
 
     /* Clean up Hash of extensions. */
@@ -718,8 +729,15 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     uv_mutex_destroy(&instance->mutex_callsite_interns);
     cleanup_callsite_interns(instance);
 
-    /* Release this interpreter's hold on Unicode database */
-    MVM_unicode_release(instance->main_thread);
+    /* Clean up Unicode hashes. */
+    for (int i = 0; i < MVM_NUM_PROPERTY_CODES; i++) {
+        MVM_uni_hash_demolish(instance->main_thread, &instance->unicode_property_values_hashes[i]);
+    }
+    MVM_free_null(instance->unicode_property_values_hashes);
+
+    MVM_uni_hash_demolish(instance->main_thread, &instance->property_codes_by_names_aliases);
+    MVM_uni_hash_demolish(instance->main_thread, &instance->property_codes_by_seq_names);
+    MVM_uni_hash_demolish(instance->main_thread, &instance->codepoints_by_name);
 
     /* Clean up spesh mutexes and close any log. */
     uv_mutex_destroy(&instance->mutex_spesh_install);
