@@ -1,11 +1,21 @@
-#include <stdlib.h>
+/* Configuration. */
+#include "gen/config.h"
+
+#if MVM_HAS_PTHREAD_SETNAME_NP
+/* pthread_setname_np only exists if we set _GNU_SOURCE extremely early.
+ * We will need to be vgilant to not accidentally use gnu extensions in
+ * other places without checking properly. */
+#ifndef _GNU_SOURCE
+#define _GNU_SOURCE
+#endif
+#endif
+
 #include <stdarg.h>
 #include <stdio.h>
 #include <setjmp.h>
 #include <stddef.h>
-
-/* Configuration. */
-#include "gen/config.h"
+#include <stdlib.h>
+#include <string.h>
 
 /* Standard integer types. */
 #include <platform/inttypes.h>
@@ -31,6 +41,9 @@
 #include <dyncall_callback.h>
 #endif
 
+/* needed in threadcontext.h */
+#include <platform/threads.h>
+
 /* forward declarations */
 #include "types.h"
 
@@ -58,6 +71,8 @@ typedef double   MVMnum64;
 #define ALIGNOF(t) ((char *)(&((struct { char c; t _h; } *)0)->_h) - (char *)0)
 #endif
 
+#define MVM_ASSERT_ALIGNED(var, align) assert(!((MVMuint64)(var) % (MVMuint64)(align)))
+
 #define MVM_ALIGN_SECTION_MASK ((MVMuint32)ALIGNOF(MVMint64) - 1)
 #define MVM_ALIGN_SECTION(offset) (((offset) + (MVM_ALIGN_SECTION_MASK)) & ~(MVM_ALIGN_SECTION_MASK))
 
@@ -82,12 +97,16 @@ typedef double   MVMnum64;
 #  define MVM_USED_BY_JIT
 #endif
 
-/* stuff for uthash */
-#define uthash_fatal(msg) MVM_exception_throw_adhoc(tc, "internal hash error: " msg)
+/* Hashes */
+#define HASH_DEBUG_ITER 0
+#define MVM_HASH_RANDOMIZE 1
+#define MVM_HASH_MAX_PROBE_DISTANCE 255
+#define MVM_HASH_INITIAL_BITS_IN_METADATA 5
 
-#include "strings/uthash_types.h"
+typedef MVMuint32 MVMHashNumItems;
+typedef MVMuint64 MVMHashv;
 
-MVM_PUBLIC const MVMint32 MVM_jit_support(void);
+MVM_PUBLIC MVMint32 MVM_jit_support(void);
 
 /* Headers for various other data structures and APIs. */
 #include "6model/6model.h"
@@ -95,13 +114,17 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "gc/debug.h"
 #include "gc/wb.h"
 #include "core/vector.h"
+#include "core/exceptions.h"
+#include "core/str_hash_table.h"
+#include "core/fixkey_hash_table.h"
+#include "core/index_hash_table.h"
+#include "core/ptr_hash_table.h"
+#include "core/uni_hash_table.h"
 #include "core/threadcontext.h"
 #include "core/instance.h"
-#include "strings/uthash.h"
 #include "core/interp.h"
 #include "core/callsite.h"
 #include "core/args.h"
-#include "core/exceptions.h"
 #include "core/alloc.h"
 #include "core/frame.h"
 #include "core/callstack.h"
@@ -125,10 +148,10 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "core/dll.h"
 #include "core/continuation.h"
 #include "debug/debugserver.h"
+#include "spesh/pea.h"
 #include "6model/reprs.h"
 #include "6model/reprconv.h"
 #include "6model/bootstrap.h"
-#include "6model/containers.h"
 #include "6model/sc.h"
 #include "6model/serialization.h"
 #include "6model/parametric.h"
@@ -143,10 +166,8 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "core/regionalloc.h"
 #include "spesh/dump.h"
 #include "spesh/debug.h"
-#include "spesh/pea.h"
 #include "spesh/graph.h"
 #include "spesh/codegen.h"
-#include "spesh/candidate.h"
 #include "spesh/manipulate.h"
 #include "spesh/args.h"
 #include "spesh/usages.h"
@@ -177,19 +198,29 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "strings/utf16.h"
 #include "strings/iter.h"
 #include "strings/ops.h"
+#include "core/fixedsizealloc.h"
+#include "io/procops.h"
+#include "core/str_hash_table_funcs.h"
+#include "core/fixkey_hash_table_funcs.h"
+#include "core/index_hash_table_funcs.h"
+#include "core/ptr_hash_table_funcs.h"
+#include "core/uni_hash_table_funcs.h"
+#include "6model/containers.h"
 #include "strings/unicode_gen.h"
 #include "strings/unicode.h"
 #include "strings/latin1.h"
 #include "strings/windows1252.h"
 #include "strings/shiftjis.h"
 #include "strings/unicode_ops.h"
+#include "strings/gb2312.h"
+#include "strings/gb18030.h"
+#include "strings/siphash/csiphash.h"
 #include "io/io.h"
 #include "io/eventloop.h"
 #include "io/syncfile.h"
 #include "io/syncsocket.h"
 #include "io/fileops.h"
 #include "io/dirops.h"
-#include "io/procops.h"
 #include "io/timers.h"
 #include "io/filewatchers.h"
 #include "io/signals.h"
@@ -197,7 +228,6 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "io/asyncsocketudp.h"
 #include "math/bigintops.h"
 #include "core/intcache.h"
-#include "core/fixedsizealloc.h"
 #include "jit/graph.h"
 #include "jit/label.h"
 #include "jit/expr.h"
@@ -220,6 +250,7 @@ MVMObject *MVM_backend_config(MVMThreadContext *tc);
 /* Top level VM API functions. */
 MVM_PUBLIC MVMInstance * MVM_vm_create_instance(void);
 MVM_PUBLIC void MVM_vm_run_file(MVMInstance *instance, const char *filename);
+MVM_PUBLIC void MVM_vm_run_bytecode(MVMInstance *instance, MVMuint8 *bytes, MVMuint32 size);
 MVM_PUBLIC void MVM_vm_dump_file(MVMInstance *instance, const char *filename);
 MVM_PUBLIC MVM_NO_RETURN void MVM_vm_exit(MVMInstance *instance) MVM_NO_RETURN_ATTRIBUTE;
 MVM_PUBLIC void MVM_vm_destroy_instance(MVMInstance *instance);
@@ -228,10 +259,18 @@ MVM_PUBLIC void MVM_vm_set_exec_name(MVMInstance *instance, const char *exec_nam
 MVM_PUBLIC void MVM_vm_set_prog_name(MVMInstance *instance, const char *prog_name);
 MVM_PUBLIC void MVM_vm_set_lib_path(MVMInstance *instance, int count, const char **lib_path);
 
+MVM_PUBLIC void MVM_vm_event_subscription_configure(MVMThreadContext *tc, MVMObject *queue, MVMObject *config);
+
 /* Returns absolute executable path. */
 MVM_PUBLIC int MVM_exepath(char* buffer, size_t* size);
 
-#if defined(__s390__)
+#ifdef _WIN32
+/* Reopens STDIN, STDOUT, STDERR to the 'NUL' device. */
+MVM_PUBLIC int MVM_set_std_handles_to_nul(void);
+#endif
+
+/* Seems that both 32 and 64 bit sparc need this crutch */
+#if defined(__s390__) || defined(__sparc__)
 AO_t AO_fetch_compare_and_swap_emulation(volatile AO_t *addr, AO_t old_val, AO_t new_val);
 # define AO_fetch_compare_and_swap_full(addr, old, newval) \
     AO_fetch_compare_and_swap_emulation(addr, old, newval)
