@@ -48,7 +48,7 @@ static void go_to_next_inline(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
         fw->inline_idx = NO_INLINE;
         return;
     }
-    MVMJitCode *jitcode = cand->jitcode;
+    MVMJitCode *jitcode = cand->body.jitcode;
     if (jitcode) {
         MVMuint32 idx = MVM_jit_code_get_active_inlines(tc, jitcode, fw->jit_position, fw->inline_idx + 1);
         if (idx < jitcode->num_inlines) {
@@ -58,8 +58,8 @@ static void go_to_next_inline(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
     }
     else {
         MVMuint32 i;
-        for (i = fw->inline_idx + 1; i < cand->num_inlines; i++) {
-            if (fw->deopt_offset > cand->inlines[i].start && fw->deopt_offset <= cand->inlines[i].end) {
+        for (i = fw->inline_idx + 1; i < cand->body.num_inlines; i++) {
+            if (fw->deopt_offset > cand->body.inlines[i].start && fw->deopt_offset <= cand->body.inlines[i].end) {
                 /* Found an applicable inline. */
                 fw->inline_idx = i;
                 return;
@@ -76,8 +76,8 @@ static void go_to_next_inline(MVMThreadContext *tc, MVMSpeshFrameWalker *fw) {
 static void go_to_first_inline(MVMThreadContext *tc, MVMSpeshFrameWalker *fw, MVMFrame *prev) {
     MVMFrame *f = fw->cur_caller_frame;
     MVMSpeshCandidate *spesh_cand = f->spesh_cand;
-    if (spesh_cand && spesh_cand->inlines) {
-        MVMJitCode *jitcode = spesh_cand->jitcode;
+    if (spesh_cand && spesh_cand->body.inlines) {
+        MVMJitCode *jitcode = spesh_cand->body.jitcode;
         if (jitcode && f->jit_entry_label) {
             void *current_position = prev && prev->extra && prev->extra->caller_jit_position
                 ? prev->extra->caller_jit_position
@@ -94,7 +94,7 @@ static void go_to_first_inline(MVMThreadContext *tc, MVMSpeshFrameWalker *fw, MV
                 ? prev->extra->caller_deopt_idx - 1
                 : MVM_spesh_deopt_find_inactive_frame_deopt_idx(tc, f);
             if (deopt_idx >= 0) {
-                fw->deopt_offset = spesh_cand->deopts[2 * deopt_idx + 1];
+                fw->deopt_offset = spesh_cand->body.deopts[2 * deopt_idx + 1];
                 fw->inline_idx = -1;
                 go_to_next_inline(tc, fw);
                 return;
@@ -169,7 +169,7 @@ MVMuint32 MVM_spesh_frame_walker_next(MVMThreadContext *tc, MVMSpeshFrameWalker
                 outer = fw->cur_caller_frame->outer;
             }
             else {
-                MVMSpeshInline *i = &(spesh_cand->inlines[fw->inline_idx]);
+                MVMSpeshInline *i = &(spesh_cand->body.inlines[fw->inline_idx]);
                 MVMCode *code = (MVMCode *)fw->cur_caller_frame->work[i->code_ref_reg].o;
                 outer = code ? code->body.outer : NULL;
             }
@@ -214,8 +214,8 @@ static void find_lex_info(MVMThreadContext *tc, MVMSpeshFrameWalker *fw, MVMFram
             *base_index_out = 0;
         }
         else {
-            *sf_out = spesh_cand->inlines[fw->inline_idx].sf;
-            *base_index_out = spesh_cand->inlines[fw->inline_idx].lexicals_start;
+            *sf_out = spesh_cand->body.inlines[fw->inline_idx].sf;
+            *base_index_out = spesh_cand->body.inlines[fw->inline_idx].lexicals_start;
         }
     }
 }
@@ -231,28 +231,22 @@ MVMuint32 MVM_spesh_frame_walker_get_lex(MVMThreadContext *tc, MVMSpeshFrameWalk
     MVMFrame *cur_frame;
     MVMStaticFrame *sf;
     MVMuint32 base_index;
-    MVMLexicalRegistry *lexical_names;
     find_lex_info(tc, fw, &cur_frame, &sf, &base_index);
-    lexical_names = sf->body.lexical_names;
-    if (lexical_names) {
-        /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
-        MVMLexicalRegistry *entry;
-        MVM_HASH_GET(tc, lexical_names, name, entry)
-        if (entry) {
-            MVMint32 index = base_index + entry->value;
-            MVMRegister *result = &cur_frame->env[index];
-            MVMuint16 kind = sf->body.lexical_types[entry->value];
-            *found_out = result;
-            *found_kind_out = kind;
-            if (vivify && kind == MVM_reg_obj && !result->o) {
-                MVMROOT(tc, cur_frame, {
+    MVMuint32 idx = MVM_get_lexical_by_name(tc, sf, name);
+    if (idx != MVM_INDEX_HASH_NOT_FOUND) {
+        MVMint32 index = base_index + idx;
+        MVMRegister *result = &cur_frame->env[index];
+        MVMuint16 kind = sf->body.lexical_types[idx];
+        *found_out = result;
+        *found_kind_out = kind;
+        if (vivify && kind == MVM_reg_obj && !result->o) {
+            MVMROOT(tc, cur_frame, {
                     MVM_frame_vivify_lexical(tc, cur_frame, index);
                 });
-            }
-            if (found_frame)
-                *found_frame = cur_frame;
-            return 1;
         }
+        if (found_frame)
+            *found_frame = cur_frame;
+        return 1;
     }
     return 0;
 }
@@ -265,7 +259,7 @@ MVMuint32 MVM_spesh_frame_walker_move_outer(MVMThreadContext *tc, MVMSpeshFrameW
         outer = fw->cur_caller_frame->outer;
     }
     else {
-        MVMSpeshInline *i = &(spesh_cand->inlines[fw->inline_idx]);
+        MVMSpeshInline *i = &(spesh_cand->body.inlines[fw->inline_idx]);
         MVMCode *code = (MVMCode *)fw->cur_caller_frame->work[i->code_ref_reg].o;
         outer = code ? code->body.outer : NULL;
     }
@@ -312,7 +306,7 @@ MVMuint32 MVM_spesh_frame_walker_move_caller_skip_thunks(MVMThreadContext *tc,
         MVMSpeshCandidate *spesh_cand = fw->cur_caller_frame->spesh_cand;
         MVMStaticFrame *sf = (fw->inline_idx == NO_INLINE || !spesh_cand)
             ? fw->cur_caller_frame->static_info
-            : spesh_cand->inlines[fw->inline_idx].sf;
+            : spesh_cand->body.inlines[fw->inline_idx].sf;
         if (!sf->body.is_thunk)
             return 1;
     }
@@ -338,83 +332,83 @@ MVMObject * MVM_spesh_frame_walker_get_lexicals_hash(MVMThreadContext *tc, MVMSp
     MVMObject *ctx_hash = MVM_repr_alloc_init(tc, hll->slurpy_hash_type);
     find_lex_info(tc, fw, &frame, &sf, &base_index);
     MVMROOT3(tc, ctx_hash, frame, sf, {
-        MVMLexicalRegistry **lexreg = sf->body.lexical_names_list;
+        MVMString **lexnames = sf->body.lexical_names_list;
         MVMuint32 i;
         for (i = 0; i < sf->body.num_lexicals; i++) {
             MVMuint16 type = sf->body.lexical_types[i];
-            MVMuint32 idx = base_index + lexreg[i]->value;
+            MVMuint32 idx = base_index + i;
             switch (type) {
                 case MVM_reg_obj: {
                     MVMObject *obj = frame->env[idx].o;
                     if (!obj)
                         obj = MVM_frame_vivify_lexical(tc, frame, idx);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, obj);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], obj);
                     break;
                 }
                 case MVM_reg_str: {
                     MVMObject *bs = MVM_repr_box_str(tc, hll->str_box_type,
                         frame->env[idx].s);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bs);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bs);
                     break;
                 }
                 case MVM_reg_int8: {
                     MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                         frame->env[idx].i8);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bi);
                     break;
                 }
                 case MVM_reg_uint8: {
                     MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                         frame->env[idx].u8);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bi);
                     break;
                 }
                 case MVM_reg_int16: {
                     MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                         frame->env[idx].i16);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bi);
                     break;
                 }
                 case MVM_reg_uint16: {
                     MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                         frame->env[idx].u16);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bi);
                     break;
                 }
                 case MVM_reg_int32: {
                     MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                         frame->env[idx].i32);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bi);
                     break;
                 }
                 case MVM_reg_uint32: {
                     MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                         frame->env[idx].u32);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bi);
                     break;
                 }
                 case MVM_reg_int64: {
                     MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                         frame->env[idx].i64);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bi);
                     break;
                 }
                 case MVM_reg_uint64: {
                     MVMObject *bi = MVM_repr_box_int(tc, hll->int_box_type,
                         frame->env[idx].u64);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bi);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bi);
                     break;
                 }
                 case MVM_reg_num32: {
                     MVMObject *bn = MVM_repr_box_num(tc, hll->num_box_type,
                         frame->env[idx].n32);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bn);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bn);
                     break;
                 }
                 case MVM_reg_num64: {
                     MVMObject *bn = MVM_repr_box_num(tc, hll->num_box_type,
                         frame->env[idx].n64);
-                    MVM_repr_bind_key_o(tc, ctx_hash, lexreg[i]->key, bn);
+                    MVM_repr_bind_key_o(tc, ctx_hash, lexnames[i], bn);
                     break;
                 }
                 default:
@@ -434,15 +428,10 @@ MVMint64 MVM_spesh_frame_walker_get_lexical_primspec(MVMThreadContext *tc,
     MVMFrame *cur_frame;
     MVMStaticFrame *sf;
     MVMuint32 base_index;
-    MVMLexicalRegistry *lexical_names;
     find_lex_info(tc, fw, &cur_frame, &sf, &base_index);
-    lexical_names = sf->body.lexical_names;
-    if (lexical_names) {
-        MVMLexicalRegistry *entry;
-        MVM_HASH_GET(tc, lexical_names, name, entry)
-        if (entry)
-            return MVM_frame_translate_to_primspec(tc, sf->body.lexical_types[entry->value]);
-    }
+    MVMuint32 idx = MVM_get_lexical_by_name(tc, sf, name);
+    if (idx != MVM_INDEX_HASH_NOT_FOUND)
+        return MVM_frame_translate_to_primspec(tc, sf->body.lexical_types[idx]);
     return -1;
 }
 
@@ -454,7 +443,7 @@ MVMObject * MVM_spesh_frame_walker_get_code(MVMThreadContext *tc, MVMSpeshFrameW
     if (fw->inline_idx == NO_INLINE || !spesh_cand)
         return fw->cur_caller_frame->code_ref;
     return fw->cur_caller_frame->work[
-        spesh_cand->inlines[fw->inline_idx].code_ref_reg
+        spesh_cand->body.inlines[fw->inline_idx].code_ref_reg
     ].o;
 }
 
@@ -464,10 +453,8 @@ MVMuint64 MVM_spesh_frame_walker_get_lexical_count(MVMThreadContext *tc, MVMSpes
     MVMFrame *cur_frame;
     MVMStaticFrame *sf;
     MVMuint32 base_index;
-    MVMLexicalRegistry *lexical_names;
     find_lex_info(tc, fw, &cur_frame, &sf, &base_index);
-    lexical_names = sf->body.lexical_names;
-    return lexical_names ? (MVMuint64)HASH_CNT(hash_handle, lexical_names) : 0;
+    return sf->body.num_lexicals;
 }
 
 /* Cleans up the spesh frame walker after use. */
